from collections import defaultdict
from collections.abc import Iterable
from typing import Any

from pdag._core import CoreModel, FunctionRelationship, ParameterABC, SubModelRelationship
from pdag.utils import merge_two_set_dicts

from .model import (
    AbsoluteParameterId,
    AbsoluteRelationshipId,
    AbsoluteStaticParameterId,
    AbsoluteStaticRelationshipId,
    AbsoluteTimeSeriesParameterId,
    AbsoluteTimeSeriesRelationshipId,
    ExecutionModel,
    FunctionRelationshipInfo,
)


def _iter_submodels_recursively(
    core_model: CoreModel,
    *,
    include_self: bool = False,
) -> Iterable[CoreModel]:
    assert core_model.is_hydrated(), "CoreModel must be hydrated."
    if include_self:
        yield core_model
    for relationship in core_model.relationships.values():
        if isinstance(relationship, SubModelRelationship):
            yield relationship.submodel
            yield from _iter_submodels_recursively(relationship.submodel)


def _iter_function_relationships_recursively(
    core_model: CoreModel,
) -> Iterable[tuple[CoreModel, FunctionRelationship[Any, Any]]]:
    """Iterate over all function relationships in the core model.

    Yields tuples of the form (submodel, relationship).
    """
    assert core_model.is_hydrated(), "CoreModel must be hydrated."
    for submodel in _iter_submodels_recursively(core_model, include_self=True):
        for relationship in submodel.relationships.values():
            if isinstance(relationship, FunctionRelationship):
                yield submodel, relationship


def _iter_submodel_relationships_recursively(
    core_model: CoreModel,
) -> Iterable[tuple[CoreModel, SubModelRelationship]]:
    """Iterate over all submodel relationships in the core model.

    Yields tuples of the form (submodel, relationship).
    """
    assert core_model.is_hydrated(), "CoreModel must be hydrated."
    for submodel in _iter_submodels_recursively(core_model, include_self=True):
        for relationship in submodel.relationships.values():
            if isinstance(relationship, SubModelRelationship):
                yield submodel, relationship


def _iter_parameters_recursively(
    core_model: CoreModel,
) -> Iterable[tuple[CoreModel, ParameterABC[Any]]]:
    """Iterate over all parameters in the core model.

    Yields tuples of the form (submodel, parameter).
    """
    assert core_model.is_hydrated(), "CoreModel must be hydrated."
    for submodel in _iter_submodels_recursively(core_model, include_self=True):
        for parameter in submodel.parameters.values():
            yield submodel, parameter


def _calculate_dependencies_of_time_series_function_relationship(
    relationship: FunctionRelationship[Any, Any],
    *,
    core_model: CoreModel,
    n_time_steps: int,
) -> tuple[
    dict[AbsoluteParameterId, set[AbsoluteRelationshipId]],  # input_parameter_id_to_relationship_ids
    dict[AbsoluteRelationshipId, set[AbsoluteParameterId]],  # relationship_id_to_output_parameter_ids
    dict[AbsoluteRelationshipId, FunctionRelationshipInfo],  # relationship_id_to_function_relationship_info
]:
    assert relationship.evaluated_at_each_time_step, (
        "This function should only be called for time series relationships."
    )

    input_parameter_id_to_relationship_ids_dd: defaultdict[AbsoluteParameterId, set[AbsoluteRelationshipId]] = (
        defaultdict(set)
    )
    relationship_id_to_output_parameter_ids_dd: defaultdict[AbsoluteRelationshipId, set[AbsoluteParameterId]] = (
        defaultdict(set)
    )
    relationship_id_to_function_relationship_info: dict[AbsoluteRelationshipId, FunctionRelationshipInfo] = {}

    match relationship.includes_past, relationship.includes_future:
        case True, True:
            msg = "Relationships with both past and future dependencies are not supported."
            raise ValueError(msg)
        case True, False:
            time_steps = list(range(1, n_time_steps))
        case False, True:
            time_steps = list(range(n_time_steps - 1))
        case False, False:
            time_steps = list(range(n_time_steps))

    for time_step in time_steps:
        relationship_id = AbsoluteTimeSeriesRelationshipId(
            model_name=core_model.name,
            name=relationship.name,
            time_step=time_step,
        )
        input_args: dict[str, AbsoluteParameterId | tuple[AbsoluteParameterId, ...]] = {}
        for input_arg_name, input_parameter_ref in relationship.inputs.items():
            input_parameter = core_model.get_parameter(input_parameter_ref)
            if input_parameter.is_time_series:
                input_parameter_id: AbsoluteParameterId = AbsoluteTimeSeriesParameterId(
                    model_name=core_model.name,
                    name=input_parameter.name,
                    time_step=time_step - 1 if input_parameter_ref.previous else time_step,
                )
            else:
                input_parameter_id = AbsoluteStaticParameterId(model_name=core_model.name, name=input_parameter.name)
            input_parameter_id_to_relationship_ids_dd[input_parameter_id].add(relationship_id)
            input_args[input_arg_name] = input_parameter_id

        output_args: list[AbsoluteParameterId | tuple[AbsoluteParameterId, ...]] = []
        for output_parameter_ref in relationship.outputs:
            output_parameter = core_model.get_parameter(output_parameter_ref)
            if output_parameter.is_time_series:
                output_parameter_id: AbsoluteParameterId = AbsoluteTimeSeriesParameterId(
                    model_name=core_model.name,
                    name=output_parameter.name,
                    time_step=time_step + 1 if output_parameter_ref.next else time_step,
                )
            else:
                output_parameter_id = AbsoluteStaticParameterId(model_name=core_model.name, name=output_parameter.name)
            relationship_id_to_output_parameter_ids_dd[relationship_id].add(output_parameter_id)
            output_args.append(output_parameter_id)

        relationship_id_to_function_relationship_info[relationship_id] = FunctionRelationshipInfo(
            function_relationship=relationship,
            input_parameter_ids=input_args,
            output_parameter_ids=tuple(output_args),
        )

    return (
        dict(input_parameter_id_to_relationship_ids_dd),
        dict(relationship_id_to_output_parameter_ids_dd),
        relationship_id_to_function_relationship_info,
    )


def _calculate_dependencies_of_static_function_relationship(
    relationship: FunctionRelationship[Any, Any],
    *,
    core_model: CoreModel,
    n_time_steps: int,
) -> tuple[
    dict[AbsoluteParameterId, set[AbsoluteRelationshipId]],  # input_parameter_id_to_relationship_ids
    dict[AbsoluteRelationshipId, set[AbsoluteParameterId]],  # relationship_id_to_output_parameter_ids
    dict[AbsoluteRelationshipId, FunctionRelationshipInfo],  # relationship_id_to_function_relationship_info
]:
    assert not relationship.evaluated_at_each_time_step, "This function should only be called for static relationships."
    relationship_id = AbsoluteStaticRelationshipId(model_name=core_model.name, name=relationship.name)
    input_parameter_ids: set[AbsoluteParameterId] = set()
    output_parameter_ids: set[AbsoluteParameterId] = set()

    input_args: dict[str, AbsoluteParameterId | tuple[AbsoluteParameterId, ...]] = {}
    for input_arg_name, input_parameter_ref in relationship.inputs.items():
        input_parameter = core_model.get_parameter(input_parameter_ref)
        if input_parameter.is_time_series:
            assert input_parameter_ref.all_time_steps, (
                "Time-series parameters for static relationships must be all-time-steps."
            )
            input_parameter_ids_local: list[AbsoluteParameterId] = [
                AbsoluteTimeSeriesParameterId(
                    model_name=core_model.name,
                    name=input_parameter.name,
                    time_step=time_step,
                )
                for time_step in range(n_time_steps)
            ]
            input_parameter_ids.update(input_parameter_ids_local)
            input_args[input_arg_name] = tuple(input_parameter_ids_local)
        else:
            input_parameter_ids.add(AbsoluteStaticParameterId(model_name=core_model.name, name=input_parameter.name))
            input_args[input_arg_name] = AbsoluteStaticParameterId(
                model_name=core_model.name,
                name=input_parameter.name,
            )

    output_args: list[AbsoluteParameterId | tuple[AbsoluteParameterId, ...]] = []
    for output_parameter_ref in relationship.outputs:
        output_parameter = core_model.get_parameter(output_parameter_ref)
        if output_parameter.is_time_series:
            match output_parameter_ref.all_time_steps, output_parameter_ref.initial:
                case True, True:
                    msg = "Time-series parameters cannot be both all-time-steps and initial."
                    raise ValueError(msg)
                case True, False:  # all-time-steps
                    output_parameter_ids_local: list[AbsoluteParameterId] = [
                        AbsoluteTimeSeriesParameterId(
                            model_name=core_model.name,
                            name=output_parameter.name,
                            time_step=time_step,
                        )
                        for time_step in range(n_time_steps)
                    ]
                    output_parameter_ids.update(output_parameter_ids_local)
                    output_args.append(tuple(output_parameter_ids_local))
                case False, True:  # initial
                    output_parameter_id_local: AbsoluteParameterId = AbsoluteTimeSeriesParameterId(
                        model_name=core_model.name,
                        name=output_parameter.name,
                        time_step=0,
                    )
                    output_parameter_ids.add(output_parameter_id_local)
                    output_args.append(output_parameter_id_local)
                case False, False:
                    msg = "Time-series parameters must be either all-time-steps or initial."
                    raise ValueError(msg)
        else:
            output_parameter_id_local = AbsoluteStaticParameterId(
                model_name=core_model.name,
                name=output_parameter.name,
            )
            output_parameter_ids.add(output_parameter_id_local)
            output_args.append(output_parameter_id_local)

    input_parameter_id_to_relationship_ids: dict[AbsoluteParameterId, set[AbsoluteRelationshipId]] = {
        input_parameter_id: {relationship_id} for input_parameter_id in input_parameter_ids
    }
    relationship_id_to_output_parameter_ids: dict[AbsoluteRelationshipId, set[AbsoluteParameterId]] = {
        relationship_id: output_parameter_ids,
    }
    relationship_id_to_function_relationship_info: dict[AbsoluteRelationshipId, FunctionRelationshipInfo] = {
        relationship_id: FunctionRelationshipInfo(
            function_relationship=relationship,
            input_parameter_ids=input_args,
            output_parameter_ids=tuple(output_args),
        ),
    }
    return (
        input_parameter_id_to_relationship_ids,
        relationship_id_to_output_parameter_ids,
        relationship_id_to_function_relationship_info,
    )


def _calculate_port_mapping_of_time_series_submodel_relationship(  # noqa: C901, PLR0912
    relationship: SubModelRelationship,
    *,
    core_model: CoreModel,
    n_time_steps: int,
) -> dict[AbsoluteParameterId, AbsoluteParameterId]:
    # parent model input to sub-model input / sub-model output to parent model input
    port_mapping: dict[AbsoluteParameterId, AbsoluteParameterId] = {}

    match relationship.includes_past, relationship.includes_future:
        case True, True:
            msg = "Relationships with both past and future dependencies are not supported."
            raise ValueError(msg)
        case True, False:
            time_steps = list(range(1, n_time_steps))
        case False, True:
            time_steps = list(range(n_time_steps - 1))
        case False, False:
            time_steps = list(range(n_time_steps))

    for time_step in time_steps:
        for input_parameter_ref_inner, input_parameter_ref_outer in relationship.inputs.items():
            input_parameter_inner = relationship.submodel.get_parameter(input_parameter_ref_inner)
            input_parameter_outer = core_model.get_parameter(input_parameter_ref_outer)
            if input_parameter_inner.is_time_series:
                input_parameter_id_inner: AbsoluteParameterId = AbsoluteTimeSeriesParameterId(
                    model_name=relationship.submodel.name,
                    name=input_parameter_inner.name,
                    time_step=time_step,
                )
            else:
                input_parameter_id_inner = AbsoluteStaticParameterId(
                    model_name=relationship.submodel.name,
                    name=input_parameter_inner.name,
                )

            if input_parameter_outer.is_time_series:
                input_parameter_id_outer: AbsoluteParameterId = AbsoluteTimeSeriesParameterId(
                    model_name=core_model.name,
                    name=input_parameter_outer.name,
                    time_step=time_step - 1 if input_parameter_ref_outer.previous else time_step,
                )
            else:
                input_parameter_id_outer = AbsoluteStaticParameterId(
                    model_name=core_model.name,
                    name=input_parameter_outer.name,
                )

            port_mapping[input_parameter_id_outer] = input_parameter_id_inner

        for output_parameter_ref_inner, output_parameter_ref_outer in relationship.outputs.items():
            output_parameter_inner = relationship.submodel.get_parameter(output_parameter_ref_inner)
            output_parameter_outer = core_model.get_parameter(output_parameter_ref_outer)
            if output_parameter_inner.is_time_series:
                output_parameter_id_inner: AbsoluteParameterId = AbsoluteTimeSeriesParameterId(
                    model_name=relationship.submodel.name,
                    name=output_parameter_inner.name,
                    time_step=time_step,
                )
            else:
                output_parameter_id_inner = AbsoluteStaticParameterId(
                    model_name=relationship.submodel.name,
                    name=output_parameter_inner.name,
                )

            if output_parameter_outer.is_time_series:
                output_parameter_id_outer: AbsoluteParameterId = AbsoluteTimeSeriesParameterId(
                    model_name=core_model.name,
                    name=output_parameter_outer.name,
                    time_step=time_step + 1 if output_parameter_ref_outer.next else time_step,
                )
            else:
                output_parameter_id_outer = AbsoluteStaticParameterId(
                    model_name=core_model.name,
                    name=output_parameter_outer.name,
                )

            port_mapping[output_parameter_id_inner] = output_parameter_id_outer

    return port_mapping


def _calculate_port_mapping_of_static_submodel_relationship(
    relationship: SubModelRelationship,
    *,
    core_model: CoreModel,
    n_time_steps: int,
) -> dict[AbsoluteParameterId, AbsoluteParameterId]:
    # parent model input to sub-model input / sub-model output to parent model input
    port_mapping: dict[AbsoluteParameterId, AbsoluteParameterId] = {}

    for input_parameter_ref_inner, input_parameter_ref_outer in relationship.inputs.items():
        input_parameter_inner = relationship.submodel.get_parameter(input_parameter_ref_inner)
        input_parameter_outer = core_model.get_parameter(input_parameter_ref_outer)
        if input_parameter_outer.is_time_series:
            assert input_parameter_ref_outer.all_time_steps, (
                "Time-series parameters for static relationships must be all-time-steps."
            )
            assert input_parameter_inner.is_time_series, "Parameter types must match."
            assert input_parameter_ref_inner.all_time_steps, "Parameter types must match."
            port_mapping.update(
                {
                    AbsoluteTimeSeriesParameterId(
                        model_name=core_model.name,
                        name=input_parameter_outer.name,
                        time_step=time_step,
                    ): AbsoluteTimeSeriesParameterId(
                        model_name=relationship.submodel.name,
                        name=input_parameter_inner.name,
                        time_step=time_step,
                    )
                    for time_step in range(n_time_steps)
                },
            )
        else:
            assert not input_parameter_outer.is_time_series, "Parameter types must match."
            port_mapping[AbsoluteStaticParameterId(model_name=core_model.name, name=input_parameter_outer.name)] = (
                AbsoluteStaticParameterId(model_name=relationship.submodel.name, name=input_parameter_inner.name)
            )

    for output_parameter_ref_inner, output_parameter_ref_outer in relationship.outputs.items():
        output_parameter_inner = relationship.submodel.get_parameter(output_parameter_ref_inner)
        output_parameter_outer = core_model.get_parameter(output_parameter_ref_outer)
        if output_parameter_outer.is_time_series:
            assert output_parameter_inner.is_time_series, "Parameter types must match."

            assert (output_parameter_ref_outer.all_time_steps and output_parameter_ref_inner.all_time_steps) or (
                output_parameter_ref_outer.initial and output_parameter_ref_inner.initial
            ), "Parameter types must match."

            time_steps = list(range(n_time_steps)) if output_parameter_ref_outer.all_time_steps else [0]

            port_mapping.update(
                {
                    AbsoluteTimeSeriesParameterId(
                        model_name=relationship.submodel.name,
                        name=output_parameter_inner.name,
                        time_step=time_step,
                    ): AbsoluteTimeSeriesParameterId(
                        model_name=core_model.name,
                        name=output_parameter_outer.name,
                        time_step=time_step,
                    )
                    for time_step in time_steps
                },
            )
        else:
            assert not output_parameter_inner.is_time_series, "Parameter types must match."
            port_mapping[
                AbsoluteStaticParameterId(model_name=relationship.submodel.name, name=output_parameter_inner.name)
            ] = AbsoluteStaticParameterId(model_name=core_model.name, name=output_parameter_outer.name)

    return port_mapping


def create_exec_model_from_core_model(
    core_model: CoreModel,
    *,
    n_time_steps: int = 1,
) -> ExecutionModel:
    parameter_ids: set[AbsoluteParameterId] = set()
    for model, parameter in _iter_parameters_recursively(core_model):
        if parameter.is_time_series:
            parameter_ids.update(
                AbsoluteTimeSeriesParameterId(
                    model_name=model.name,
                    name=parameter.name,
                    time_step=time_step,
                )
                for time_step in range(n_time_steps)
            )
        else:
            parameter_ids.add(AbsoluteStaticParameterId(model_name=model.name, name=parameter.name))

    relationships: dict[AbsoluteRelationshipId, FunctionRelationshipInfo] = {}
    input_parameter_id_to_relationship_ids: dict[AbsoluteParameterId, set[AbsoluteRelationshipId]] = {}
    relationship_id_to_output_parameter_ids: dict[AbsoluteRelationshipId, set[AbsoluteParameterId]] = {}
    port_mapping: dict[AbsoluteParameterId, AbsoluteParameterId] = {}

    for model, function_relationship in _iter_function_relationships_recursively(core_model):
        if function_relationship.evaluated_at_each_time_step:
            (
                input_parameter_id_to_relationship_ids_local,
                relationship_id_to_output_parameter_ids_local,
                relationships_local,
            ) = _calculate_dependencies_of_time_series_function_relationship(
                function_relationship,
                core_model=model,
                n_time_steps=n_time_steps,
            )
        else:
            (
                input_parameter_id_to_relationship_ids_local,
                relationship_id_to_output_parameter_ids_local,
                relationships_local,
            ) = _calculate_dependencies_of_static_function_relationship(
                function_relationship,
                core_model=model,
                n_time_steps=n_time_steps,
            )
        relationships.update(relationships_local)

        input_parameter_id_to_relationship_ids = merge_two_set_dicts(
            input_parameter_id_to_relationship_ids,
            input_parameter_id_to_relationship_ids_local,
        )
        relationship_id_to_output_parameter_ids = merge_two_set_dicts(
            relationship_id_to_output_parameter_ids,
            relationship_id_to_output_parameter_ids_local,
        )

    for model, sub_model_relationship in _iter_submodel_relationships_recursively(core_model):
        if sub_model_relationship.evaluated_at_each_time_step:
            port_mapping_local = _calculate_port_mapping_of_time_series_submodel_relationship(
                sub_model_relationship,
                core_model=model,
                n_time_steps=n_time_steps,
            )
        else:
            port_mapping_local = _calculate_port_mapping_of_static_submodel_relationship(
                sub_model_relationship,
                core_model=model,
                n_time_steps=n_time_steps,
            )
        port_mapping.update(port_mapping_local)

    return ExecutionModel(
        parameter_ids=parameter_ids,
        relationship_infos=relationships,
        input_parameter_id_to_relationship_ids=input_parameter_id_to_relationship_ids,
        relationship_id_to_output_parameter_ids=relationship_id_to_output_parameter_ids,
        port_mapping=port_mapping,
    )



from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable
from dataclasses import dataclass, field
from typing import Any, ClassVar, Self

from .utils import InitArgsRecorder


@dataclass(frozen=True)  # Frozen to be valid as a dictionary key
class ParameterRef(InitArgsRecorder):
    name: str
    previous: bool = field(default=False, kw_only=True)
    next: bool = field(default=False, kw_only=True)
    # Options to access time series data
    # TODO: Make this more flexible
    initial: bool = field(default=False, kw_only=True)
    all_time_steps: bool = field(default=False, kw_only=True)
    __init_args__: tuple[Any, ...] = field(init=False, compare=False, hash=False)
    __init_kwargs__: dict[str, Any] = field(init=False, compare=False, hash=False)

    def __post_init__(self) -> None:
        if sum([self.previous, self.next, self.initial, self.all_time_steps]) > 1:
            msg = "Parameter reference cannot have more than one of previous, next, initial, or all_time_steps set."
            raise ValueError(msg)
        if not self.name.isidentifier():
            msg = "Parameter reference name must be a valid identifier."
            raise ValueError(msg)

    @property
    def normal(self) -> bool:
        return not any([self.previous, self.next, self.initial, self.all_time_steps])

    def __str__(self) -> str:
        s = self.name
        if self.previous:
            s += ".previous"
        if self.next:
            s += ".next"
        if self.initial:
            s += ".initial"
        if self.all_time_steps:
            s += ".all_time_steps"
        return s

    @classmethod
    def from_str(cls, s: str) -> Self:
        if ".previous" in s:
            name, _ = s.split(".previous")
            return cls(name=name, previous=True)
        if ".next" in s:
            name, _ = s.split(".next")
            return cls(name=name, next=True)
        if ".initial" in s:
            name, _ = s.split(".initial")
            return cls(name=name, initial=True)
        if ".all_time_steps" in s:
            name, _ = s.split(".all_time_steps")
            return cls(name=name, all_time_steps=True)
        return cls(name=s)


@dataclass
class ParameterABC[T](InitArgsRecorder, ABC):
    type: ClassVar[str] = "base"
    name: str
    is_time_series: bool = field(default=False, kw_only=True)

    @abstractmethod
    def is_hydrated(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def get_type_hint(self) -> str:
        raise NotImplementedError


@dataclass
class RealParameter(ParameterABC[float]):
    type: ClassVar[str] = "real"
    unit: str | None = None
    lower_bound: float | None = None
    upper_bound: float | None = None

    def is_hydrated(self) -> bool:
        return True

    def get_type_hint(self) -> str:
        return "float"


@dataclass
class CategoricalParameter(ParameterABC[str]):
    type: ClassVar[str] = "categorical"
    categories: set[str]

    def is_hydrated(self) -> bool:
        return True

    def get_type_hint(self) -> str:
        return f"Literal[{', '.join(repr(category) for category in self.categories)}]"


@dataclass
class ParameterCollectionABC(ABC, InitArgsRecorder):
    type: ClassVar[str] = "collection"
    name: str

    @abstractmethod
    def iter_parameter_names(self) -> Iterable[str]:
        raise NotImplementedError

    @abstractmethod
    def is_hydrated(self) -> bool:
        raise NotImplementedError


@dataclass
class ParameterMapping(ParameterCollectionABC):
    type: ClassVar[str] = "mapping"
    name: str
    mapping: dict[str, str]

    def iter_parameter_names(self) -> Iterable[str]:
        return self.mapping.values()

    def is_hydrated(self) -> bool:
        return True


type StrArrayOrStr = list["StrArrayOrStr"] | str


@dataclass
class ParameterArray(ParameterCollectionABC):
    type: ClassVar[str] = "array"
    name: str
    shape: tuple[int, ...]
    array: list[StrArrayOrStr]

    def iter_parameter_names(self) -> Iterable[str]:
        def _iter_parameter_names(array_or_str: StrArrayOrStr) -> Iterable[str]:
            if isinstance(array_or_str, str):
                yield array_or_str
            else:
                for element in array_or_str:
                    yield from _iter_parameter_names(element)

        for element in self.array:
            yield from _iter_parameter_names(element)

    def is_hydrated(self) -> bool:
        return True


@dataclass
class RelationshipABC(ABC, InitArgsRecorder):
    type: ClassVar[str] = "relationship"
    name: str
    evaluated_at_each_time_step: bool = field(default=False, kw_only=True)

    @abstractmethod
    def is_hydrated(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def iter_input_parameter_refs(self) -> Iterable[ParameterRef]:
        raise NotImplementedError

    @abstractmethod
    def iter_output_parameter_refs(self) -> Iterable[ParameterRef]:
        raise NotImplementedError

    @property
    def includes_past(self) -> bool:
        return any(param_ref.previous for param_ref in self.iter_input_parameter_refs())

    @property
    def includes_future(self) -> bool:
        return any(param_ref.next for param_ref in self.iter_output_parameter_refs())


@dataclass
class FunctionRelationship[**P, T](RelationshipABC):
    type: ClassVar[str] = "function"
    name: str
    inputs: dict[str, ParameterRef]
    outputs: list[ParameterRef]
    function_body: str
    output_is_scalar: bool = field(kw_only=True)
    _function: Callable[P, T] | None = field(default=None, compare=False, kw_only=True)
    evaluated_at_each_time_step: bool = field(default=False, kw_only=True)

    def is_hydrated(self) -> bool:
        return self._function is not None

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        if self._function is None:
            msg = f"Function relationship {self.name} has not been hydrated."
            raise ValueError(msg)
        return self._function(*args, **kwargs)

    def iter_input_parameter_refs(self) -> Iterable[ParameterRef]:
        return self.inputs.values()

    def iter_output_parameter_refs(self) -> Iterable[ParameterRef]:
        return self.outputs


@dataclass
class SubModelRelationship(RelationshipABC):
    type: ClassVar[str] = "submodel"
    name: str
    submodel_name: str
    inputs: dict[ParameterRef, ParameterRef]  # sub-model parameter ref -> parent model parameter ref
    outputs: dict[ParameterRef, ParameterRef]  # sub-model parameter ref -> parent model parameter ref
    _submodel: "CoreModel | None" = field(default=None, compare=False, kw_only=True)
    evaluated_at_each_time_step: bool = field(default=False, kw_only=True)

    def is_hydrated(self) -> bool:
        return self._submodel is not None

    def iter_input_parameter_refs(self) -> Iterable[ParameterRef]:
        return self.inputs.values()

    def iter_output_parameter_refs(self) -> Iterable[ParameterRef]:
        return self.outputs.values()

    @property
    def submodel(self) -> "CoreModel":
        if self._submodel is None:
            msg = f"Sub-model relationship {self.name} has not been hydrated."
            raise ValueError(msg)
        return self._submodel


@dataclass
class CoreModel:
    name: str
    parameters: dict[str, ParameterABC[Any]]
    collections: dict[str, ParameterCollectionABC]
    relationships: dict[str, RelationshipABC]

    def is_hydrated(self) -> bool:
        return (
            all(parameter.is_hydrated() for parameter in self.parameters.values())
            and all(collection.is_hydrated() for collection in self.collections.values())
            and all(relationship.is_hydrated() for relationship in self.relationships.values())
        )

    def is_dynamic(self) -> bool:
        return any(parameter.is_time_series for parameter in self.parameters.values())

    def get_parameter(self, parameter_ref: ParameterRef) -> ParameterABC[Any]:
        return self.parameters[parameter_ref.name]


@dataclass
class Module:
    pre_models: str
    models: list[CoreModel]  # Assuming topological order
    post_models: str

    def is_hydrated(self) -> bool:
        return all(model.is_hydrated() for model in self.models)



"""pdag: A package for defining and working with directed acyclic graphs (DAGs) of parameters and relationships."""

__all__ = [
    "AbsoluteParameterId",
    "AbsoluteRelationshipId",
    "AbsoluteStaticParameterId",
    "AbsoluteStaticRelationshipId",
    "AbsoluteTimeSeriesParameterId",
    "AbsoluteTimeSeriesRelationshipId",
    "CategoricalParameter",
    "CoreModel",
    "ExecutionModel",
    "FunctionRelationship",
    "Model",
    "NodeId",
    "ParameterABC",
    "ParameterArray",
    "ParameterCollectionABC",
    "ParameterMapping",
    "ParameterRef",
    "RealParameter",
    "RelationshipABC",
    "SubModelRelationship",
    "core_model_to_content",
    "core_model_to_dataclass_notation_ast",
    "create_exec_model_from_core_model",
    "execute_exec_model",
    "module_to_content",
    "relationship",
    "utils",
]


from . import utils  # expose utils module
from ._core import (
    CategoricalParameter,
    CoreModel,
    FunctionRelationship,
    ParameterABC,
    ParameterArray,
    ParameterCollectionABC,
    ParameterMapping,
    ParameterRef,
    RealParameter,
    RelationshipABC,
    SubModelRelationship,
)
from ._exec import (
    AbsoluteParameterId,
    AbsoluteRelationshipId,
    AbsoluteStaticParameterId,
    AbsoluteStaticRelationshipId,
    AbsoluteTimeSeriesParameterId,
    AbsoluteTimeSeriesRelationshipId,
    ExecutionModel,
    NodeId,
    create_exec_model_from_core_model,
    execute_exec_model,
)
from ._notation import (
    Model,
    core_model_to_content,
    core_model_to_dataclass_notation_ast,
    module_to_content,
    relationship,
)



import ast

from rich import print

class_def = '''\
# Some code before the class definition
import pdag

class SquareRootModel(pdag.Model):
    """A model for square root."""

    x = pdag.RealParameter("x")
    y = pdag.RealParameter("y")
    z = pdag.CategoricalParameter("z", categories={"pos", "neg"})

    @pdag.relationship
    @staticmethod
    def sqrt(
        *,
        x_arg: Annotated[float, pdag.ParameterRef("x")],
        z_arg: Annotated[Literal["neg", "pos"], pdag.ParameterRef("z")],
    ) -> Annotated[float, pdag.ParameterRef("y")]:
        if z_arg == "pos":
            return float(x_arg**0.5)

# Some code after the class definition
x = 0
'''

tree = ast.parse(class_def)
print(ast.dump(tree, indent=2))



__all__ = [
    "Model",
    "ParameterRef",
    "core_model_to_content",
    "core_model_to_dataclass_notation_ast",
    "module_to_content",
    "relationship",
]
from ._decorators import relationship
from .construct import core_model_to_content, core_model_to_dataclass_notation_ast, module_to_content
from .model import Model



from typing import Annotated

import numpy as np

import pdag._experimental as pdag


class C:
    @staticmethod
    def sqrt(x: Annotated[float, pdag.RealParameter()]) -> float:
        return float(np.sqrt(x))


if __name__ == "__main__":
    import inspect

    from rich import print

    function_body = inspect.getsource(C.sqrt)
    print(function_body)
    source_lines = inspect.getsourcelines(C.sqrt)
    print(source_lines)



import ast

import asttokens

source = '''# Some code before the class definition
import pdag

class SquareRootModel(pdag.Model):
    """A model for square root."""

    x: pdag.RealParameter("x")
    y: pdag.RealParameter("y")
    z: pdag.CategoricalParameter("z", categories={"pos", "neg"})
    w: pdag.RealParameter("w")

    @pdag.relationship
    @staticmethod
    def sqrt(
        *,
        x_arg: float,
        z_arg: str
    ) -> tuple:
        return x_arg ** 0.5

# Some code after the class definition
x = 0
'''

# Parse the source with asttokens for accurate token positions.
at = asttokens.ASTTokens(source, parse=True)
# Find the class definition node by name.
class_node = next(node for node in at.tree.body if isinstance(node, ast.ClassDef) and node.name == "SquareRootModel")

# Determine the starting and ending positions.
start_line = class_node.first_token.start[0]
end_line = class_node.last_token.end[0]

# Split the code by lines.
lines = source.splitlines()
pre_classdef = "\n".join(lines[: start_line - 1])
classdef = "\n".join(lines[start_line - 1 : end_line])
post_classdef = "\n".join(lines[end_line:])

print("PRE-CLASSDEF:\n", pre_classdef)
print("CLASSDEF:\n", classdef)
print("POST-CLASSDEF:\n", post_classdef)



class SquareRootModel(pdag.Model):
    x: Annotated[float, pdag.RealParameter('x')]
    y: Annotated[float, pdag.RealParameter('y')]
    z: Annotated[Literal['neg', 'pos'], pdag.CategoricalParameter('z', categories=frozenset({'neg', 'pos'}))]

    @pdag.relationship
    @staticmethod
    def sqrt(*, x_arg: Annotated[float, pdag.Parameter('x')], z_arg: Annotated[Literal['neg', 'pos'], pdag.Parameter('z')]) -> Annotated[float, pdag.Parameter('y')]:
        if z_arg == 'pos':
            return float(x_arg ** 0.5)


from dataclasses import dataclass

from pdag.utils import InitArgsRecorder


@dataclass
class C(InitArgsRecorder):
    x: int
    y: str = "default"


@dataclass
class D(C):
    z: float = 0.0


# When the user initializes the classes:
c = C(10)
d = D(10, "hello", 3.14)
d2 = D(42, z=3.14)

print("c was initialized with:", c.__init_args__, c.__init_kwargs__)
print("d was initialized with:", d.__init_args__, d.__init_kwargs__)
print("d2 was initialized with:", d2.__init_args__, d2.__init_kwargs__)



from time import sleep


def f():
    print("Hello, World! I am f.")
    sleep(1)
    return g()


def g():
    print("Hello, World! I am g.")
    sleep(1)
    return f()


if __name__ == "__main__":
    g()



"""Poliynomials example to demonstrate how to include a model in another model."""

import pdag
from pdag.examples.polynomials import PolynomialModel


def test_model_name() -> None:
    assert PolynomialModel.name == "PolynomialModel"


def test_polynomial_model() -> None:
    core_model = PolynomialModel.to_core_model()
    exec_model = pdag.create_exec_model_from_core_model(core_model)
    results = pdag.execute_exec_model(
        exec_model,
        inputs={
            pdag.AbsoluteStaticParameterId("PolynomialModel", "a0"): 1.0,
            pdag.AbsoluteStaticParameterId("PolynomialModel", "a1"): 2.0,
            pdag.AbsoluteStaticParameterId("PolynomialModel", "a2"): 3.0,
            pdag.AbsoluteStaticParameterId("PolynomialModel", "x"): 4.0,
        },
    )

    assert results == {
        pdag.AbsoluteStaticParameterId("PolynomialModel", "a0"): 1.0,
        pdag.AbsoluteStaticParameterId("PolynomialModel", "a1"): 2.0,
        pdag.AbsoluteStaticParameterId("PolynomialModel", "a2"): 3.0,
        pdag.AbsoluteStaticParameterId("PolynomialModel", "x"): 4.0,
        pdag.AbsoluteStaticParameterId("PolynomialModel", "x_squared"): 4**2,
        pdag.AbsoluteStaticParameterId("PolynomialModel", "y"): 1 + 2 * 4 + 3 * 4**2,
        pdag.AbsoluteStaticParameterId("SquareModel", "x"): 4.0,
        pdag.AbsoluteStaticParameterId("SquareModel", "y"): 4**2,
    }



from collections import defaultdict
from dataclasses import dataclass, field
from typing import Any

from pdag._core import FunctionRelationship


@dataclass(frozen=True, slots=True)
class AbsoluteStaticParameterId:
    model_name: str
    name: str


@dataclass(frozen=True, slots=True)
class AbsoluteTimeSeriesParameterId:
    model_name: str
    name: str
    time_step: int


@dataclass(frozen=True, slots=True)
class AbsoluteStaticRelationshipId:
    model_name: str
    name: str


@dataclass(frozen=True, slots=True)
class AbsoluteTimeSeriesRelationshipId:
    model_name: str
    name: str
    time_step: int


# Type aliases
type AbsoluteParameterId = AbsoluteStaticParameterId | AbsoluteTimeSeriesParameterId
type AbsoluteRelationshipId = AbsoluteStaticRelationshipId | AbsoluteTimeSeriesRelationshipId
type NodeId = AbsoluteParameterId | AbsoluteRelationshipId


@dataclass(slots=True)
class FunctionRelationshipInfo:
    function_relationship: FunctionRelationship[Any, Any]
    input_parameter_ids: dict[str, AbsoluteParameterId | tuple[AbsoluteParameterId, ...]]
    output_parameter_ids: tuple[AbsoluteParameterId | tuple[AbsoluteParameterId, ...], ...]


@dataclass
class ExecutionModel:
    parameter_ids: set[AbsoluteParameterId]
    # SubModelRelationships should be flattened into FunctionRelationships
    relationship_infos: dict[AbsoluteRelationshipId, FunctionRelationshipInfo]
    input_parameter_id_to_relationship_ids: dict[AbsoluteParameterId, set[AbsoluteRelationshipId]]
    relationship_id_to_output_parameter_ids: dict[AbsoluteRelationshipId, set[AbsoluteParameterId]]
    # parent model output to sub-model input / sub-model output to parent model input
    port_mapping: dict[AbsoluteParameterId, AbsoluteParameterId]

    # Derived attributes
    relationship_id_to_input_parameter_ids: dict[AbsoluteRelationshipId, set[AbsoluteParameterId]] = field(
        init=False,
        repr=False,
        compare=False,
    )
    output_parameter_id_to_relationship_ids: dict[AbsoluteParameterId, set[AbsoluteRelationshipId]] = field(
        init=False,
        repr=False,
        compare=False,
    )
    port_mapping_inverse: dict[AbsoluteParameterId, AbsoluteParameterId] = field(init=False, repr=False, compare=False)

    def __post_init__(self) -> None:
        relationship_id_to_input_parameter_ids_dd: defaultdict[AbsoluteRelationshipId, set[AbsoluteParameterId]] = (
            defaultdict(set)
        )
        for input_parameter_id, relationship_ids in self.input_parameter_id_to_relationship_ids.items():
            for relationship_id in relationship_ids:
                relationship_id_to_input_parameter_ids_dd[relationship_id].add(input_parameter_id)
        self.relationship_id_to_input_parameter_ids = dict(relationship_id_to_input_parameter_ids_dd)

        output_parameter_id_to_relationship_ids_dd: defaultdict[AbsoluteParameterId, set[AbsoluteRelationshipId]] = (
            defaultdict(set)
        )
        for relationship_id, output_parameter_ids in self.relationship_id_to_output_parameter_ids.items():
            for output_parameter_id in output_parameter_ids:
                output_parameter_id_to_relationship_ids_dd[output_parameter_id].add(relationship_id)
        self.port_mapping_inverse = {value: key for key, value in self.port_mapping.items()}



"""Utility functions and classes for the pdag package."""

import ast
import inspect
from collections import defaultdict, deque
from collections.abc import Callable, Collection, Hashable, Mapping
from textwrap import dedent
from typing import Any, Self

import asttokens


class InitArgsRecorder:
    """A mixin class that records the arguments used to initialize the instance."""

    # These are not class variables, but instance variables
    # They are declared here to inform type checkers.
    __init_args__: tuple[Any, ...]
    __init_kwargs__: dict[str, Any]

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        """Record the arguments used to initialize the instance."""
        # Create the instance first
        instance = super().__new__(cls)
        # Save the arguments for later inspection
        # This is a bit of a hack to avoid assigning to attributes
        # even if the inherited dataclass has frozen=True
        object.__setattr__(instance, "__init_args__", args)
        object.__setattr__(instance, "__init_kwargs__", kwargs)
        return instance

    def get_init_args(self) -> tuple[Any, ...]:
        """Return the arguments used to initialize the instance."""
        return self.__init_args__

    def get_init_kwargs(self) -> dict[str, Any]:
        """Return the keyword arguments used to initialize the instance."""
        return self.__init_kwargs__


def get_function_body(func: Callable[..., Any]) -> str:
    """Get the body of a function as a string."""
    source = inspect.getsource(func)
    source = dedent(source)
    # Parse the source with asttokens to keep track of positions.
    atok = asttokens.ASTTokens(source, parse=True)

    # Locate the first FunctionDef node in the AST.
    function_node = None
    for node in ast.walk(atok.tree):  # type: ignore[arg-type]
        if isinstance(node, ast.FunctionDef):
            function_node = node
            break
    if function_node is None:
        msg = "Function definition not found in source."
        raise ValueError(msg)

    # Extract the source for each statement in the function body.
    start = atok.get_text_range(function_node.body[0], padded=True)[0]
    end = atok.get_text_range(function_node.body[-1], padded=True)[1]
    body = dedent(source[start:end])
    if not body.endswith("\n"):
        body += "\n"
    return body


def topological_sort[T: Hashable](dependencies: Mapping[T, Collection[T]]) -> list[T]:
    """Sort a graph of dependencies topologically.

    Given a dependency graph (a dict mapping a node to a collection of nodes that depend on it),
    perform a topological sort and return an ordered list of nodes.
    Raises an error if a cycle is detected.
    """
    indegree: defaultdict[T, int] = defaultdict(int)
    for node, deps in dependencies.items():
        indegree[node] = indegree.get(node, 0)
        for dep in deps:
            indegree[dep] += 1
    q = deque([node for node, deg in indegree.items() if deg == 0])
    order = []
    while q:
        node = q.popleft()
        order.append(node)
        for dep in dependencies.get(node, []):
            indegree[dep] -= 1
            if indegree[dep] == 0:
                q.append(dep)
    if len(order) != len(indegree):
        msg = "Cycle detected in relationship dependencies!"
        raise ValueError(msg)
    return order


def merge_two_set_dicts[K, V](d1: Mapping[K, set[V]], d2: Mapping[K, set[V]]) -> dict[K, set[V]]:
    """Merge two dictionaries where each value is a set.

    Values for duplicate keys are merged (unioned).
    """
    result = dict(d1)
    for key, value in d2.items():
        result[key] = result.get(key, set()) | value
    return result



import inspect
from abc import ABCMeta
from collections.abc import Callable, Mapping
from typing import Any, ClassVar, cast, get_args, get_origin

from typing_extensions import _AnnotatedAlias

from pdag._core import (
    CoreModel,
    FunctionRelationship,
    ParameterABC,
    ParameterCollectionABC,
    ParameterRef,
    RelationshipABC,
    SubModelRelationship,
)
from pdag.utils import get_function_body


def _get_inputs_from_signature(sig: inspect.Signature) -> dict[str, ParameterRef]:
    def _get_param_ref_from_annotation(annotations: _AnnotatedAlias) -> ParameterRef:
        args = get_args(annotations)
        return next(iter(arg for arg in args if isinstance(arg, ParameterRef)))

    return {param.name: _get_param_ref_from_annotation(param.annotation) for param in sig.parameters.values()}


def _get_outputs_from_signature(sig: inspect.Signature) -> tuple[list[ParameterRef], bool]:
    def single_annotation_to_parameter_ref(annotations: _AnnotatedAlias) -> ParameterRef:
        args = get_args(annotations)
        return next(iter(arg for arg in args if isinstance(arg, ParameterRef)))

    if get_origin(sig.return_annotation) is not tuple:
        parameter_name = single_annotation_to_parameter_ref(sig.return_annotation)
        return [parameter_name], True
    args = get_args(sig.return_annotation)
    return [single_annotation_to_parameter_ref(arg) for arg in args], False


def function_to_function_relationship[**P, T](func: Callable[P, T]) -> FunctionRelationship[P, T]:
    # Get the function's signature
    sig = inspect.signature(func)
    inputs = _get_inputs_from_signature(sig)
    outputs, output_is_scalar = _get_outputs_from_signature(sig)
    function_body = get_function_body(func)
    return FunctionRelationship(
        name=func.__name__,
        inputs=inputs,
        outputs=outputs,
        function_body=function_body,
        output_is_scalar=output_is_scalar,
        _function=func,
    )


class ModelMeta(ABCMeta):
    def __new__(metacls, name: str, bases: tuple[type[Any], ...], namespace: dict[str, Any]) -> type:  # noqa: N804
        cls = super().__new__(metacls, name, bases, namespace)

        # Tell type checker that cls is essentially of type Model
        cls = cast(type["Model"], cls)

        cls.name = name
        cls.__pdag_collections__ = {}
        cls.__pdag_parameters__ = {
            parameter_name: parameter
            for parameter_name, parameter in namespace.items()
            if isinstance(parameter, ParameterABC)
        }

        # TODO: Check validity of relationship definition
        cls.__pdag_relationships__ = {}
        for relationship_name, relationship in namespace.items():
            if isinstance(relationship, RelationshipABC):
                for output_parameter_ref in relationship.iter_output_parameter_refs():
                    output_parameter = cls.__pdag_parameters__[output_parameter_ref.name]
                    if output_parameter.is_time_series and (output_parameter_ref.normal or output_parameter_ref.next):
                        relationship.evaluated_at_each_time_step = True
                        break
                cls.__pdag_relationships__[relationship_name] = relationship

        return cls


class Model(metaclass=ModelMeta):
    name: ClassVar[str]
    __pdag_parameters__: dict[str, ParameterABC[Any]]
    __pdag_collections__: dict[str, ParameterCollectionABC]
    __pdag_relationships__: dict[str, RelationshipABC]

    @classmethod
    def parameters(cls) -> dict[str, ParameterABC[Any]]:
        return cls.__pdag_parameters__

    @classmethod
    def collections(cls) -> dict[str, ParameterCollectionABC]:
        return {}  # TODO: Implement

    @classmethod
    def relationships(cls) -> dict[str, RelationshipABC]:
        return cls.__pdag_relationships__

    @classmethod
    def to_core_model(cls) -> CoreModel:
        return CoreModel(
            name=cls.__name__,
            parameters=cls.parameters(),
            collections=cls.collections(),
            relationships=cls.relationships(),
        )

    @classmethod
    def to_relationship(
        cls,
        /,
        name: str,
        *,
        inputs: Mapping[ParameterRef, ParameterRef],
        outputs: Mapping[ParameterRef, ParameterRef],
    ) -> SubModelRelationship:
        return SubModelRelationship(
            name=name,
            submodel_name=cls.name,
            inputs=dict(inputs),
            outputs=dict(outputs),
            _submodel=cls.to_core_model(),
        )



import ast
from collections.abc import Mapping
from typing import Any

from pdag._core import (
    CoreModel,
    FunctionRelationship,
    Module,
    ParameterABC,
    ParameterRef,
    RelationshipABC,
    SubModelRelationship,
)


def _object_to_ast_value(obj: Any) -> ast.expr:
    """Convert an object to an AST value.

    This relies on the `repr` of the object being a valid Python expression.
    """
    return ast.parse(repr(obj)).body[0].value  # type: ignore[no-any-return,attr-defined]


def _parameter_nodes_to_ast_statements(nodes: Mapping[str, ParameterABC[Any]]) -> list[ast.stmt]:
    return [
        ast.Assign(
            targets=[ast.Name(id=name, ctx=ast.Store())],
            value=ast.Call(
                func=ast.Attribute(
                    value=ast.Name(id="pdag", ctx=ast.Load()),
                    attr=parameter.__class__.__name__,
                    ctx=ast.Load(),
                ),
                args=[_object_to_ast_value(arg) for arg in parameter.get_init_args()],
                keywords=[
                    ast.keyword(arg=key, value=_object_to_ast_value(value))
                    for key, value in parameter.get_init_kwargs().items()
                ],
            ),
        )
        for name, parameter in nodes.items()
    ]


def _type_hint_to_ast_node(type_hint: str) -> ast.expr:
    return ast.parse(type_hint).body[0].value  # type: ignore[no-any-return,attr-defined]


def _make_annotated_type_hint(parameter: ParameterABC[Any], parameter_ref: ParameterRef) -> ast.Subscript:
    return ast.Subscript(
        value=ast.Name(id="Annotated", ctx=ast.Load()),
        slice=ast.Tuple(
            elts=[
                _type_hint_to_ast_node(parameter.get_type_hint()),
                ast.Call(
                    func=ast.Attribute(
                        value=ast.Name(id="pdag", ctx=ast.Load()),
                        attr="ParameterRef",
                        ctx=ast.Load(),
                    ),
                    args=[_object_to_ast_value(arg) for arg in parameter_ref.get_init_args()],
                    keywords=[
                        ast.keyword(arg=key, value=_object_to_ast_value(value))
                        for key, value in parameter_ref.get_init_kwargs().items()
                    ],
                ),
            ],
            ctx=ast.Load(),
        ),
        ctx=ast.Load(),
    )


def _make_arg_ast(arg_name: str, parameter: ParameterABC[Any], parameter_ref: ParameterRef) -> ast.arg:
    return ast.arg(
        arg=arg_name,
        annotation=_make_annotated_type_hint(parameter, parameter_ref),
    )


def _make_function_body_ast(function_body: str) -> list[ast.stmt]:
    return ast.parse(function_body).body


def _make_return_ast(
    parameters: list[ParameterABC[Any]],
    parameter_refs: list[ParameterRef],
    *,
    force_tuple: bool = False,
) -> ast.Subscript:
    assert len(parameters) == len(parameter_refs)
    if len(parameters) == 1 and not force_tuple:
        return _make_annotated_type_hint(parameters[0], parameter_refs[0])
    return ast.Subscript(
        value=ast.Name(id="tuple", ctx=ast.Load()),
        slice=ast.Tuple(
            elts=[
                _make_annotated_type_hint(parameter, parameter_ref)
                for parameter, parameter_ref in zip(parameters, parameter_refs, strict=True)
            ],
            ctx=ast.Load(),
        ),
    )


def _function_relationship_to_function_def(
    relationship: FunctionRelationship[Any, Any],
    *,
    core_model: CoreModel,
) -> ast.FunctionDef:
    return ast.FunctionDef(
        name=relationship.name,
        args=ast.arguments(
            kwonlyargs=[
                _make_arg_ast(arg_name, core_model.parameters[parameter_ref.name], parameter_ref)
                for arg_name, parameter_ref in relationship.inputs.items()
            ],
            kw_defaults=[None for _ in relationship.inputs],
        ),
        body=_make_function_body_ast(relationship.function_body),
        decorator_list=[
            ast.Attribute(
                value=ast.Name(id="pdag", ctx=ast.Load()),
                attr="relationship",
                ctx=ast.Load(),
            ),
            ast.Name(id="staticmethod", ctx=ast.Load()),
        ],
        returns=_make_return_ast(
            [core_model.parameters[parameter_ref.name] for parameter_ref in relationship.outputs],
            relationship.outputs,
            force_tuple=not relationship.output_is_scalar,
        ),
    )


def _submodel_relationship_to_function_def(
    relationship: SubModelRelationship,
    *,
    core_model: CoreModel,
) -> ast.FunctionDef:
    raise NotImplementedError


def _relationship_to_function_def(
    relationship: RelationshipABC,
    *,
    core_model: CoreModel,
) -> ast.FunctionDef:
    if isinstance(relationship, FunctionRelationship):
        return _function_relationship_to_function_def(relationship, core_model=core_model)
    if isinstance(relationship, SubModelRelationship):
        return _submodel_relationship_to_function_def(relationship, core_model=core_model)
    msg = f"Unknown relationship type: {type(relationship)}"
    raise ValueError(msg)


def core_model_to_dataclass_notation_ast(core_model: CoreModel) -> ast.ClassDef:
    parameters = _parameter_nodes_to_ast_statements(core_model.parameters)
    relationships: list[ast.stmt] = [
        _relationship_to_function_def(relationship, core_model=core_model)
        for relationship in core_model.relationships.values()
    ]
    return ast.ClassDef(
        name=core_model.name,
        bases=[
            ast.Attribute(
                value=ast.Name(id="pdag", ctx=ast.Load()),
                attr="Model",
                ctx=ast.Load(),
            ),
        ],
        keywords=[],
        body=parameters + relationships,
    )


def core_model_to_content(core_model: CoreModel) -> str:
    module = ast.Module(body=[core_model_to_dataclass_notation_ast(core_model)], type_ignores=[])
    module = ast.fix_missing_locations(module)
    return ast.unparse(module)


def module_to_content(module: Module) -> str:
    return module.pre_models + "\n\n".join(core_model_to_content(model) for model in module.models) + module.post_models



import ast
import inspect
from textwrap import dedent
from typing import Annotated, Literal

import pytest

import pdag


class SquareRootModel(pdag.Model):
    x = pdag.RealParameter("x")
    y = pdag.RealParameter("y")
    z = pdag.CategoricalParameter("z", categories={"pos", "neg"})

    @pdag.relationship
    @staticmethod
    def sqrt(
        *,
        x_arg: Annotated[float, pdag.ParameterRef("x")],
        z_arg: Annotated[Literal["pos", "neg"], pdag.ParameterRef("z")],
    ) -> Annotated[float, pdag.ParameterRef("y")]:
        if z_arg == "pos":
            return float(x_arg**0.5)
        return -float(x_arg**0.5)


square_root_core_model = pdag.CoreModel(
    name="SquareRootModel",
    parameters={
        "x": pdag.RealParameter("x"),
        "y": pdag.RealParameter("y"),
        "z": pdag.CategoricalParameter("z", categories={"pos", "neg"}),
    },
    collections={},
    relationships={
        "sqrt": pdag.FunctionRelationship(
            name="sqrt",
            inputs={"x_arg": pdag.ParameterRef("x"), "z_arg": pdag.ParameterRef("z")},
            outputs=[pdag.ParameterRef("y")],
            output_is_scalar=True,
            function_body=dedent(
                """\
                if z_arg == "pos":
                    return float(x_arg**0.5)
                return -float(x_arg**0.5)
                """,
            ),
            evaluated_at_each_time_step=False,
        ),
    },
)

CASES = [
    (SquareRootModel, square_root_core_model),
]


@pytest.mark.parametrize(("dc_notation", "core_model"), CASES)
def test_core_model_to_dataclass_notation_by_ast(dc_notation: type[pdag.Model], core_model: pdag.CoreModel) -> None:
    dc_notation_source = dedent(inspect.getsource(dc_notation))

    # Compare AST
    class_def_constructed = pdag.core_model_to_dataclass_notation_ast(core_model)
    expected = ast.parse(dc_notation_source).body[0]
    assert ast.dump(class_def_constructed, indent=2) == ast.dump(expected, indent=2)


@pytest.mark.parametrize(("dc_notation", "core_model"), CASES)
def test_core_model_to_dataclass_notation_by_source(dc_notation: type[pdag.Model], core_model: pdag.CoreModel) -> None:
    """This test is redundant, as it is essentially the same as test_core_model_to_dataclass_notation_by_ast.

    But it helps debug the conversion to source code.
    """
    dc_notation_source = dedent(inspect.getsource(dc_notation))

    # Compare unparsed source code
    source_constructed = pdag.core_model_to_content(core_model)
    assert source_constructed == ast.unparse(ast.parse(dc_notation_source))


@pytest.mark.parametrize(("dc_notation", "core_model"), CASES)
def test_dataclass_notation_to_core_model(
    dc_notation: type[pdag.Model],
    core_model: pdag.CoreModel,
) -> None:
    # Compare CoreModel
    exported_core_model = dc_notation.to_core_model()
    assert exported_core_model == core_model



"""Example models for building expansion."""

from dataclasses import dataclass
from typing import ClassVar

import pdag


class NpvCalculationModel(pdag.Model):
    """A model for calculating the net present value of a project."""

    @dataclass
    class Settings:
        """Settings that are specific to the model instances."""

        n_time_steps: int


class BuildingExpansionModel(pdag.Model):
    """A model for building expansion."""

    class Constants:
        """Constants that are shared across all model instances."""

        building_states: ClassVar[tuple[str, ...]] = ("none", "opt-33", "opt-57", "exp-33", "exp-57")
        actions: ClassVar[tuple[str, ...]] = (
            "none",
            "build-opt-33",
            "build-opt-57",
            "teardown-opt-33-and-build-opt-57",
            "build-exp-33",
            "exp-to-57",
        )
        policy_types: ClassVar[tuple[str, ...]] = (
            "none",
            "build-opt-57",
            "build-opt-33-and-rebuild",
            "build-exp-33-and-expand",
        )

    @dataclass
    class Settings:
        """Settings that are specific to the model instances."""

        n_time_steps: int = 4
        revenue_per_floor: float = 3



from typing import Annotated, Literal

from rich import print

import pdag


class SquareRootModel(pdag.Model):
    """A model for square root."""

    x: Annotated[float, pdag.RealParameter("x")]
    y: Annotated[float, pdag.RealParameter("y")]
    z: Annotated[Literal["neg", "pos"], pdag.CategoricalParameter("z", categories=frozenset({"pos", "neg"}))]
    w: Annotated[float, pdag.RealParameter("w")]

    @pdag.relationship
    @staticmethod
    def sqrt(
        *,
        x_arg: Annotated[float, pdag.ParameterRef("x")],
        z_arg: Annotated[Literal["neg", "pos"], pdag.ParameterRef("z")],
    ) -> Annotated[float, pdag.ParameterRef("y")]:
        if z_arg == "pos":
            print("===")
            return float(x_arg**0.5)
        return -float(x_arg**0.5)


core_model = SquareRootModel.to_core_model()
print(core_model)
assert core_model.is_hydrated()



from collections import defaultdict, deque
from collections.abc import Mapping
from typing import Any, cast

from pdag.utils import topological_sort

from .model import (
    AbsoluteParameterId,
    AbsoluteStaticParameterId,
    AbsoluteStaticRelationshipId,
    AbsoluteTimeSeriesParameterId,
    AbsoluteTimeSeriesRelationshipId,
    ExecutionModel,
    NodeId,
)


def execute_exec_model(  # noqa: C901, PLR0912
    exec_model: ExecutionModel,
    inputs: Mapping[AbsoluteParameterId, Any],
) -> dict[AbsoluteParameterId, Any]:
    # TODO: Exec only part of the model
    # Calculate dependencies
    dependencies_dd: defaultdict[NodeId, set[NodeId]] = defaultdict(
        set,
        cast(dict[NodeId, set[NodeId]], exec_model.input_parameter_id_to_relationship_ids)
        | cast(dict[NodeId, set[NodeId]], exec_model.relationship_id_to_output_parameter_ids),
    )
    for src, dest in exec_model.port_mapping.items():
        dependencies_dd[src].add(dest)
    dependencies = dict(dependencies_dd)
    # Sort nodes topologically
    sorted_nodes = topological_sort(dependencies)

    sorted_nodes_queue = deque(sorted_nodes)
    results: dict[AbsoluteParameterId, Any] = {}
    while sorted_nodes_queue:
        node_id = sorted_nodes_queue.popleft()

        if node_id in results:
            continue

        if isinstance(node_id, AbsoluteStaticParameterId | AbsoluteTimeSeriesParameterId):
            if node_id in inputs:
                results[node_id] = inputs[node_id]
                continue
            if node_id in exec_model.port_mapping_inverse:
                parent_node_id = exec_model.port_mapping_inverse[node_id]
                if parent_node_id in inputs:
                    results[node_id] = inputs[parent_node_id]
                    continue
                if parent_node_id in results:
                    results[node_id] = results[parent_node_id]
                    continue
                msg = f"Parent node {parent_node_id} not in results"
                raise ValueError(msg)
            msg = f"Node {node_id} not in inputs or port_mapping"
            raise ValueError(msg)

        if isinstance(node_id, AbsoluteStaticRelationshipId | AbsoluteTimeSeriesRelationshipId):
            relationship_info = exec_model.relationship_infos[node_id]
            input_values: dict[str, Any] = {}
            for input_arg_name, input_parameter_id in relationship_info.input_parameter_ids.items():
                if isinstance(input_parameter_id, tuple):
                    input_values[input_arg_name] = tuple(
                        results[input_param_id] for input_param_id in input_parameter_id
                    )
                else:
                    input_values[input_arg_name] = results[input_parameter_id]

            output_values = relationship_info.function_relationship(**input_values)

            if relationship_info.function_relationship.output_is_scalar:
                output_values = (output_values,)
            for output_parameter_id, output_value in zip(
                relationship_info.output_parameter_ids,
                output_values,
                strict=True,
            ):
                if isinstance(output_parameter_id, tuple):
                    results.update(zip(output_parameter_id, output_value, strict=True))
                else:
                    results[output_parameter_id] = output_value

    return results



__all__ = [
    "AbsoluteParameterId",
    "AbsoluteRelationshipId",
    "AbsoluteStaticParameterId",
    "AbsoluteStaticRelationshipId",
    "AbsoluteTimeSeriesParameterId",
    "AbsoluteTimeSeriesRelationshipId",
    "ExecutionModel",
    "NodeId",
    "create_exec_model_from_core_model",
    "execute_exec_model",
]
from .core import execute_exec_model
from .model import (
    AbsoluteParameterId,
    AbsoluteRelationshipId,
    AbsoluteStaticParameterId,
    AbsoluteStaticRelationshipId,
    AbsoluteTimeSeriesParameterId,
    AbsoluteTimeSeriesRelationshipId,
    ExecutionModel,
    NodeId,
)
from .to_exec_model import create_exec_model_from_core_model



from collections.abc import Callable

from pdag._core import FunctionRelationship

from .model import function_to_function_relationship


def relationship[**P, T](func: Callable[P, T]) -> FunctionRelationship[P, T]:
    """Decorate a function to mark it as a relationship."""
    return function_to_function_relationship(func)



"""Poliynomials example to demonstrate how to include a model in another model."""

from typing import Annotated

import pdag


class SquareModel(pdag.Model):
    """Square model."""

    x = pdag.RealParameter("x")
    y = pdag.RealParameter("y")

    @pdag.relationship
    @staticmethod
    def square(*, x: Annotated[float, pdag.ParameterRef("x")]) -> Annotated[float, pdag.ParameterRef("y")]:  # noqa: D102
        return x**2


class PolynomialModel(pdag.Model):
    """Polynomial model."""

    a0 = pdag.RealParameter("a0")
    a1 = pdag.RealParameter("a1")
    a2 = pdag.RealParameter("a2")
    x = pdag.RealParameter("x")
    x_squared = pdag.RealParameter("x_squared")
    y = pdag.RealParameter("y")

    calc_square_term = SquareModel.to_relationship(
        "calc_square_term",
        inputs={pdag.ParameterRef("x"): pdag.ParameterRef("x")},
        outputs={pdag.ParameterRef("y"): pdag.ParameterRef("x_squared")},
    )

    @pdag.relationship
    @staticmethod
    def polynomial(  # noqa: D102
        *,
        a0: Annotated[float, pdag.ParameterRef("a0")],
        a1: Annotated[float, pdag.ParameterRef("a1")],
        a2: Annotated[float, pdag.ParameterRef("a2")],
        x: Annotated[float, pdag.ParameterRef("x")],
        x_squared: Annotated[float, pdag.ParameterRef("x_squared")],
    ) -> Annotated[float, pdag.ParameterRef("y")]:
        return a0 + a1 * x + a2 * x_squared


if __name__ == "__main__":
    from rich import print  # noqa: A004

    core_model = PolynomialModel.to_core_model()
    print(core_model)
    exec_model = pdag.create_exec_model_from_core_model(core_model)
    print(exec_model)



"""Diamond MDP model example."""

import pdag
from pdag.examples.diamond_mdp import DiamondMdpModel


def test_model_name() -> None:
    assert DiamondMdpModel.name == "DiamondMdpModel"


def test_diamond_mdp() -> None:
    core_model = DiamondMdpModel.to_core_model()
    exec_model = pdag.create_exec_model_from_core_model(core_model, n_time_steps=4)
    results = pdag.execute_exec_model(
        exec_model,
        inputs={
            pdag.AbsoluteStaticParameterId("DiamondMdpModel", "policy"): "left",
            pdag.AbsoluteTimeSeriesParameterId("DiamondMdpModel", "location", 0): "start",
        },
    )
    assert results == {
        pdag.AbsoluteStaticParameterId("DiamondMdpModel", "policy"): "left",
        pdag.AbsoluteTimeSeriesParameterId("DiamondMdpModel", "location", time_step=0): "start",
        pdag.AbsoluteTimeSeriesParameterId("DiamondMdpModel", "reward", time_step=0): 0.0,
        pdag.AbsoluteTimeSeriesParameterId("DiamondMdpModel", "action", time_step=0): "go_left",
        pdag.AbsoluteTimeSeriesParameterId("DiamondMdpModel", "location", time_step=1): "left",
        pdag.AbsoluteTimeSeriesParameterId("DiamondMdpModel", "reward", time_step=1): 0.0,
        pdag.AbsoluteTimeSeriesParameterId("DiamondMdpModel", "action", time_step=1): "move_forward",
        pdag.AbsoluteTimeSeriesParameterId("DiamondMdpModel", "location", time_step=2): "end",
        pdag.AbsoluteTimeSeriesParameterId("DiamondMdpModel", "action", time_step=2): "none",
        pdag.AbsoluteTimeSeriesParameterId("DiamondMdpModel", "reward", time_step=2): 1.0,
        pdag.AbsoluteTimeSeriesParameterId("DiamondMdpModel", "location", time_step=3): "end",
        pdag.AbsoluteTimeSeriesParameterId("DiamondMdpModel", "action", time_step=3): "none",
        pdag.AbsoluteTimeSeriesParameterId("DiamondMdpModel", "reward", time_step=3): 0.0,
        pdag.AbsoluteStaticParameterId("DiamondMdpModel", "cumulative_reward"): 1.0,
    }



"""Diamond MDP model example."""

from typing import Annotated, Literal

import pdag


class DiamondMdpModel(pdag.Model):
    """Diamond MDP model."""

    policy = pdag.CategoricalParameter("policy", categories={"left", "right"})
    location = pdag.CategoricalParameter("location", categories={"start", "left", "right", "end"}, is_time_series=True)
    action = pdag.CategoricalParameter(
        "action",
        categories={"go_left", "go_right", "move_forward", "none"},
        is_time_series=True,
    )
    # Initial value of the reward is not calculated in the model
    reward = pdag.RealParameter("reward", is_time_series=True)
    cumulative_reward = pdag.RealParameter("cumulative_reward")

    @pdag.relationship
    @staticmethod
    def action_selection(  # noqa: D102
        *,
        policy: Annotated[Literal["left", "right"], pdag.ParameterRef("policy")],
        location: Annotated[Literal["start", "left", "right", "end"], pdag.ParameterRef("location")],
    ) -> Annotated[Literal["go_left", "go_right", "move_forward", "none"], pdag.ParameterRef("action")]:
        match location, policy:
            case "start", "left":
                return "go_left"
            case "start", "right":
                return "go_right"
            case "left", "left":
                return "move_forward"
            case "right", "right":
                return "move_forward"
            case "end", _:
                return "none"

        msg = "Invalid policy and location combination"
        raise ValueError(msg)

    @pdag.relationship
    @staticmethod
    def state_transition(  # noqa: D102
        *,
        location: Annotated[Literal["start", "left", "right", "end"], pdag.ParameterRef("location")],
        action: Annotated[Literal["go_left", "go_right", "move_forward", "none"], pdag.ParameterRef("action")],
    ) -> Annotated[Literal["start", "left", "right", "end"], pdag.ParameterRef("location", next=True)]:
        match location, action:
            case "start", "go_left":
                return "left"
            case "start", "go_right":
                return "right"
            case "left", "move_forward":
                return "end"
            case "right", "move_forward":
                return "end"
            case "end", _:
                return "end"

        msg = "Invalid location and action combination"
        raise ValueError(msg)

    @pdag.relationship
    @staticmethod
    def initial_reward() -> Annotated[float, pdag.ParameterRef("reward", initial=True)]:  # noqa: D102
        return 0.0

    @pdag.relationship
    @staticmethod
    def reward_function(  # noqa: D102
        *,
        previous_location: Annotated[
            Literal["start", "left", "right", "end"],
            pdag.ParameterRef("location", previous=True),
        ],
        action: Annotated[  # noqa: ARG004
            Literal["go_left", "go_right", "move_forward", "none"],
            pdag.ParameterRef("action", previous=True),
        ],
        location: Annotated[Literal["start", "left", "right", "end"], pdag.ParameterRef("location")],
    ) -> Annotated[float, pdag.ParameterRef("reward")]:
        if previous_location != "end" and location == "end":
            return 1.0
        return 0.0

    @pdag.relationship
    @staticmethod
    def cumulative_reward_calculation(  # noqa: D102
        *,
        reward: Annotated[list[float], pdag.ParameterRef("reward", all_time_steps=True)],
    ) -> Annotated[float, pdag.ParameterRef("cumulative_reward")]:
        return sum(reward)


if __name__ == "__main__":
    from rich import print  # noqa: A004

    core_model = DiamondMdpModel.to_core_model()
    print(core_model)
    exec_model = pdag.create_exec_model_from_core_model(core_model, n_time_steps=4)
    print(exec_model)
    results = pdag.execute_exec_model(
        exec_model,
        inputs={
            pdag.AbsoluteStaticParameterId("DiamondMdpModel", "policy"): "left",
            pdag.AbsoluteTimeSeriesParameterId("DiamondMdpModel", "location", 0): "start",
        },
    )

    print(results)



from typing import Annotated, Literal

from rich import print

import pdag


class SquareRootModel(pdag.Model):
    """A model for square root."""

    x = pdag.RealParameter("x")
    y = pdag.RealParameter("y")
    z = pdag.CategoricalParameter("z", categories={"pos", "neg"})
    w = pdag.RealParameter("w")

    @pdag.relationship
    @staticmethod
    def sqrt(
        *,
        x_arg: Annotated[float, pdag.ParameterRef("x")],
        z_arg: Annotated[Literal["neg", "pos"], pdag.ParameterRef("z")],
    ) -> Annotated[float, pdag.ParameterRef("y")]:
        if z_arg == "pos":
            return float(x_arg**0.5)
        return -float(x_arg**0.5)


print(SquareRootModel.to_core_model())
import inspect

source = inspect.getsource(SquareRootModel)
print(repr(source))



import ast
from pathlib import Path
from textwrap import dedent

import pdag

output_path = Path(__file__).parent / "model.py"

x = pdag.RealParameter("x")

core_model = pdag.CoreModel(
    name="SquareRootModel",
    parameters={
        "x": pdag.RealParameter("x"),
        "y": pdag.RealParameter("y"),
        "z": pdag.CategoricalParameter("z", categories={"pos", "neg"}),
    },
    collections={},
    relationships={
        "sqrt": pdag.FunctionRelationship(
            name="sqrt",
            inputs={"x_arg": "x", "z_arg": "z"},
            outputs=["y"],
            output_is_scalar=True,
            function_body=dedent(
                """\
                if z_arg == 'pos':
                    return float(x_arg**0.5)
                return -float(x_arg**0.5)
                """,
            ),
        ),
    },
)

print(core_model)

class_def = pdag.core_model_to_dataclass_notation_ast(core_model)

print(ast.dump(class_def, indent=2))
# Wrap the class definition in a Module node
module = ast.Module(body=[class_def], type_ignores=[])

# Fix missing location info (lineno, col_offset, etc.)
module = ast.fix_missing_locations(module)

# Convert the AST back to source code (requires Python 3.9+)
source_code = ast.unparse(module)

print(source_code)

# Write the generated source code to a file
with output_path.open("w") as f:
    f.write(source_code)



import ast
from textwrap import dedent

from rich import print


def strip_staticmethod_decorator(function_body: str) -> str:
    lines = function_body.split("\n")
    if lines[0].strip() == "@staticmethod":
        lines = lines[1:]
    return "\n".join(lines)


# function_lines = [
#     "    @staticmethod\n",
#     "    def sqrt(x: Annotated[float, pdag.RealParameter()]) -> float:\n",
#     "        return float(np.sqrt(x))\n",
# ]

function_lines = [
    "    @staticmethod\n",
    "    def square(x):\n",
    "        return x * x\n",
]

function_body = dedent("".join(function_lines))
print(function_body)
tree = ast.parse(function_body)
print(ast.dump(tree, indent=2))

namespace = {}
exec(function_body, namespace)

print(namespace)
function_restored = namespace["square"]
print(function_restored(3))
